#define MAX_HOSTLEN 16
#define MAX_DOMLEN 16
#define MAX_USERLEN 16
#define RESP_LEN 24
#define NONCE_LEN 8

struct ntlm_msg1 {
    unsigned char protocol[8];
    unsigned char type;         /* 1 */
    unsigned char zero1[3];
    unsigned char flags1;
    unsigned char flags2;
    unsigned char zero2[2];

    unsigned char dom_len[4];
    unsigned char dom_off[4];

    unsigned char host_len[4];
    unsigned char host_off[4];

#if 0
    unsigned char data[0];
#endif
} __attribute__((packed));

struct ntlm_msg2 {
    unsigned char protocol[8];
    unsigned char type;         /* 2 */
    unsigned char zero1[7];
    unsigned char msg_len[4];
    unsigned char flags[2];
    unsigned char zero2[2];

    unsigned char nonce[8];
    unsigned char zero3[8];
} __attribute__((packed));

struct ntlm_msg3 {
    unsigned char protocol[8];
    unsigned char type;         /* 3 */
    unsigned char zero1[3];

    unsigned char lm_len[4];
    unsigned char lm_off[4];

    unsigned char nt_len[4];
    unsigned char nt_off[4];

    unsigned char dom_len[4];
    unsigned char dom_off[4];

    unsigned char user_len[4];
    unsigned char user_off[4];

    unsigned char host_len[4];
    unsigned char host_off[4];

    unsigned char msg_len[4];

#if 0
    unsigned char data[0];
#endif
} __attribute__((packed));

typedef struct ntlmssp_info {
    int msg_type;
    unsigned char user[MAX_USERLEN + 1];
    unsigned char host[MAX_HOSTLEN + 1];
    unsigned char domain[MAX_DOMLEN + 1];
    unsigned char lm[RESP_LEN];
    unsigned char nt[RESP_LEN];
} ntlmssp_info_rec;

#define little_endian_word(x) x[0] + (((unsigned)x[1]) << 8)
#define set_little_endian_word(x,y) x[0]=(y&0xff);x[1]=(y>>8)

static int 
ntlm_msg_type(unsigned char *raw_msg, unsigned msglen)
{
    struct ntlm_msg1 *msg = (struct ntlm_msg1 *) raw_msg;

    if (msglen < 9)
        return -1;
    if (strncmp(msg->protocol, "NTLMSSP", 8))
        return -1;
    return msg->type;
}

static int 
ntlm_extract_mem(request_rec * r, unsigned char *dst,
                 unsigned char *src, unsigned srclen,
                 unsigned char *off, unsigned char *len,
                 unsigned max)
{
    unsigned o = little_endian_word(off);
    unsigned l = little_endian_word(len);
    if (l > max)
        return -1;
    if (o >= srclen)
        return -1;
    if (o + l > srclen)
        return -1;
    src += o;
    while (l-- > 0)
        *dst++ = *src++;
    return 0;
}

static int 
ntlm_extract_string(request_rec * r, unsigned char *dst,
                    unsigned char *src, unsigned srclen,
                    unsigned char *off, unsigned char *len,
                    unsigned max)
{
    unsigned o = little_endian_word(off);
    unsigned l = little_endian_word(len);
    if (l > max)
        return -1;
    if (o >= srclen)
        return -1;
    if (o + l > srclen)
        return -1;
    src += o;
    while (l-- > 0)
        *dst++ = *src++;
    *dst = 0;
    return 0;
}

static int 
ntlm_extract_unicode(request_rec * r, unsigned char *dst,
                     unsigned char *src, unsigned srclen,
                     unsigned char *off, unsigned char *len,
                     unsigned max)
{
    unsigned o = little_endian_word(off);
    unsigned l = little_endian_word(len) / 2;   /* Unicode! */
    if (l > max)
        return -1;
    if (o >= srclen)
        return -1;
    if (o + l > srclen)
        return -1;
    src += o;
    while (l > 0) {
        /* Unicode to ASCII */
        *dst++ = *src;
        src += 2;
        l -= 2;
    }
    *dst = 0;
    return 0;
}

static int 
ntlm_msg1_gethostname(request_rec * r, unsigned char *raw_msg,
                      unsigned msglen, unsigned char *hostname)
{
    struct ntlm_msg1 *msg = (struct ntlm_msg1 *) raw_msg;
    if (ntlm_extract_string(r, hostname, (char *) msg, msglen,
                            msg->host_off, msg->host_len, MAX_HOSTLEN))
        return 1;
    return 0;
}

static int 
ntlm_msg1_getdomainname(request_rec * r, unsigned char *raw_msg,
                        unsigned msglen, unsigned char *domainname)
{
    struct ntlm_msg1 *msg = (struct ntlm_msg1 *) raw_msg;
    if (ntlm_extract_string(r, domainname, (char *) msg,
                            msglen, msg->dom_off, msg->dom_len, MAX_DOMLEN))
        return 2;
    return 0;
}

static int 
ntlm_msg3_getlm(request_rec * r, unsigned char *raw_msg, unsigned msglen,
                unsigned char *lm)
{
    struct ntlm_msg3 *msg = (struct ntlm_msg3 *) raw_msg;
    if (ntlm_extract_mem(r, lm, (char *) msg, msglen, msg->lm_off,
                         msg->lm_len, RESP_LEN))
        return 4;
    return 0;
}

static int 
ntlm_msg3_getnt(request_rec * r, unsigned char *raw_msg, unsigned msglen,
                unsigned char *nt)
{
    struct ntlm_msg3 *msg = (struct ntlm_msg3 *) raw_msg;
    if (ntlm_extract_mem(r, nt, (char *) msg, msglen, msg->nt_off,
                         msg->nt_len, RESP_LEN))
        return 8;
    return 0;
}

static int 
ntlm_msg3_getusername(request_rec * r, unsigned char *raw_msg,
                      unsigned msglen, unsigned char *username)
{
    struct ntlm_msg3 *msg = (struct ntlm_msg3 *) raw_msg;
    if (ntlm_extract_unicode(r, username, (char *) msg, msglen,
                             msg->user_off, msg->user_len, MAX_USERLEN))
        return 16;
    return 0;
}

static int 
ntlm_msg3_gethostname(request_rec * r, unsigned char *raw_msg, unsigned msglen,
                      unsigned char *hostname)
{
    struct ntlm_msg3 *msg = (struct ntlm_msg3 *) raw_msg;
    if (ntlm_extract_unicode(r, hostname, (char *) msg, msglen,
                             msg->host_off, msg->host_len, MAX_HOSTLEN))
        return 0;  /* this one FAILS, but since the value is not used,
                    * we just pretend it was ok. */
    return 0;
}

static int 
ntlm_msg3_getdomainname(request_rec * r, unsigned char *raw_msg,
                        unsigned msglen, unsigned char *domainname)
{
    struct ntlm_msg3 *msg = (struct ntlm_msg3 *) raw_msg;
    if (ntlm_extract_unicode(r, domainname, (char *) msg, msglen,
                             msg->dom_off, msg->dom_len, MAX_DOMLEN))
        return 64;
    return 0;
}

static int 
ntlm_decode_msg(request_rec * r, struct ntlmssp_info *info,
                unsigned char *raw_msg, unsigned msglen)
{
    switch (info->msg_type = ntlm_msg_type(raw_msg, msglen)) {
      case 1:
          return ntlm_msg1_gethostname(r, raw_msg, msglen, info->host)
              + ntlm_msg1_getdomainname(r, raw_msg, msglen, info->domain);
      case 3:
          return ntlm_msg3_getlm(r, raw_msg, msglen, info->lm)
              + ntlm_msg3_getnt(r, raw_msg, msglen, info->nt)
              + ntlm_msg3_getusername(r, raw_msg, msglen, info->user)
              + ntlm_msg3_gethostname(r, raw_msg, msglen, info->host)
              + ntlm_msg3_getdomainname(r, raw_msg, msglen, info->domain);
    }
    return -1;
}

static int 
ntlm_encode_msg2(unsigned char *nonce, struct ntlm_msg2 *msg)
{
    memset(msg, 0, sizeof(struct ntlm_msg2));
    strcpy(msg->protocol, "NTLMSSP");
    msg->type = 0x02;
    set_little_endian_word(msg->msg_len, sizeof(struct ntlm_msg2));
    set_little_endian_word(msg->flags, 0x8201);
    memcpy(msg->nonce, nonce, sizeof(msg->nonce));
    return 0;
}
